// MakePtr is a macro that allows you to easily add to values (including
// pointers) together without dealing with C's pointer arithmetic.  It
// essentially treats the last two parameters as DWORDs.  The first
// parameter is used to typecast the result to the appropriate pointer type.
#ifndef SYMBOLREAD_H
#define SYMBOLREAD_H
#define MakePtr( cast, ptr, addValue ) (cast)( (DWORD)(ptr) + (addValue) )

void DumpHeader(PIMAGE_FILE_HEADER pImageFileHeader);
void DumpOptionalHeader(PIMAGE_OPTIONAL_HEADER pImageOptionalHeader);
void DumpSectionTable(PIMAGE_SECTION_HEADER section,
					  unsigned cSections,
					  BOOL IsEXE);
BOOL LookupSymbolName(DWORD index, PSTR buffer, UINT length);
void DumpSymbolTable(PIMAGE_SYMBOL pSymbolTable, unsigned cSymbols);
LPVOID GetSectionPtr(PSTR name, PIMAGE_NT_HEADERS pNTHeader, DWORD imageBase);
PIMAGE_SECTION_HEADER GetSectionHeader(PSTR name, PIMAGE_NT_HEADERS pNTHeader);
void DumpRawSectionData(PIMAGE_SECTION_HEADER section,
						PVOID base,
						unsigned cSections);
void DumpLineNumbers(PIMAGE_LINENUMBER pln, DWORD count, DWORD offset);
void HexDump(PBYTE ptr, DWORD length);
int FindExeLineNrs(unsigned int idx);
int FindRelocation(DWORD addr, char *name, PIMAGE_SECTION_HEADER pSection, int range);
int FindRelocationInInstruction(int addr,int range,char *where);
#pragma pack(1)
#define S_COMPILE	0x1
#define S_UDT       0x4
#define S_START     0x5
#define S_END       0x6
#define S_OBJNAME   0x9
#define S_BPREL32	0x200
#define S_LDATA32	0x201
#define S_GDATA32	0x202
#define S_GPROC32	0x205
#define S_THUNK32	0x206
#define S_PUB32		0x203
#define S_LPROC32	0x204
#define S_BLOCK32	0x207
#define S_LABEL32	0x209
#define S_PROCREF   0x400
#define S_DATAREF   0x401
#define S_REGISTER 2
#define S_ALIGN		0x402
typedef struct tagGenericList {
	struct tagGenericList *Next;
	char *data;
} GENERIC_LIST;

struct procInfo {
	long pParent;
	long pEnd;
	long pNext;
	long procLength;
	long debugStart;
	long debugEnd;
	long offset;
	short segment;
	short procType;
	char flags;
	union {
		char fname[4];
		char *Name;
	} u;
};

struct blockInfo {
	long pParent;
	long pEnd;
	long length;
	long offset;
	short segment;
	char name[1];
};
struct localDataInfo {
	long offset;
	short segment;
	short type;
	char name[1];
};
struct autoDataInfo {
	long offset;	/* offset from BP */
	short type;
	char name[1];
};

struct UDT {	/* User Defined Type */
	short type;
	char name[1];
};

struct codeLabel {
	long offset;
	short segment;
	char flags;
	char name[1];
};


typedef struct tagSeg {
	struct tagSeg *Next;
	char *Name;
	int Index;
	unsigned long StartOffset;
	unsigned long EndOffset;
	int LineNrCount;
	int *LinesOffsets;
	int *LinesNrs;
	GENERIC_LIST *LocalProcedures;
	GENERIC_LIST *GlobalProcedures;
} SEGMENT_DEBUG_INFO;

extern SEGMENT_DEBUG_INFO *dbgInfoRoot;
extern int NumberOfSymbols;
int DumpCodeViewSymbols(PBYTE ptr, DWORD length,SEGMENT_DEBUG_INFO *sdi,int test);
void DumpCodeViewTypes(PBYTE ptr, DWORD length,int checkSignature,long *lp);

/* All values not defined are reserved for future use.*/
#define sstModule	0x120
#define sstTypes	0x121
#define sstPublic	0x122
#define sstPublicSym	0x123
#define sstSymbols	0x124
#define sstAlignSym	0x125
#define sstSrcLnSeg	0x126
#define sstSrcModule	0x127
#define sstLibraries	0x128
#define sstGlobalSym	0x129
#define sstGlobalPub	0x12a
#define sstGlobalTypes	0x12b
#define sstMPC	0x12c
#define sstSegMap	0x12d
#define sstSegName	0x12e
#define sstPreComp	0x12f
/* unused	0x130 */
#define sstOffsetMap16	0x131
#define sstOffsetMap32	0x132
#define sstFileIndex	0x133
#define sstStaticSym	0x134

#if 0
#define FRAME_FPO 0
#define FRAME_TRAP 1
#define FRAME_TSS 2
typedef struct _FOP_DATA {
	DWORD ulOffStart; /* Start of the function */
	DWORD cbProcSize; /* size of the function */
	DWORD Locals;	  /* Number of bytes in locals */
	WORD  Params;     /* Number of bytes in params */
	WORD  cbProlog:8; /* Number of bytes of prologue code */
	WORD  cbRegs  :3; /* Number of registers saved */
	WORD  fHasSEH :1; /* True if SEH (?) in function */
	WORD  fUseBP  :1; /* True if ebp has been allocated */
	WORD  reserved:1;
	WORD  cbFrame :2; /* frame type */
}FPO_DATA;
#endif
#pragma pack(4)
#endif

