#include <windows.h>
#ifdef WINELIB
#include <unix.h>
#include <peexe.h>
extern char *ctime(long *);
#else
extern char *ctime(int *);
#endif
#include <stdio.h>
#include "common.h"
#include "exedump.h"
#include "extrnvar.h"
#include <stdlib.h>
#include <string.h>
void DisplayLengthPrefixedString(char *p);
typedef PIMAGE_COFF_SYMBOLS_HEADER PIMAGE_DEBUG_INFO;
SEGMENT_DEBUG_INFO *dbgInfoRoot = NULL;
PIMAGE_DEBUG_INFO PCOFFDebugInfo = 0;
static PIMAGE_SECTION_HEADER resSection;
char *SzDebugFormats[] = {
"UNKNOWN/BORLAND", "COFF", "CODEVIEW", "FPO", "MISC", "EXCEPTION", "FIXUP"};
struct subsection {
    short subsection;
    short iMod;
    long ifo;
    long cb;
};
struct tagsstModule {
    short ovlNumber;
    short iLib;
    short cSeg;
    short Style;
    struct SegInfo *SegInfoTable;
    char *name;
};
struct SegInfo {
    short Seg;
    short pad;
    long offset;
    long cbSeg;
};
/*
(0x0120) sstModule
This describes the basic information about an object module including code segments,
module name, and the number of segments for the modules that follow.  Directory
entries for sstModules precede all other subsection directory entries.
2           2       2       2       *       *
ovlNumber   iLib    cSeg    Style   SegInfo Name
ovlNumber   Overlay number
iLib    Index into sstLibraries subsection if this module was linked from a library
cSeg    Count of the number of code segments this module contributes to
Style   Debugging style for this module.  Currently only "CV" is defined.
        A module can have only one debugging style.
        If a module contains debugging information in an unrecognized style,the
        information will be discarded.
SegInfo Detailed information about each segment that code is contributed to.
        This is an array of cSeg count segment information descriptor structures.
Name    Length-prefixed name of module
SegInfo is a  structure that describes each segment to which a module contributes code.
        It is formatted as follows:
        2   2   4       4
        Seg pad offset  cbSeg
        Seg Segment that this structure describes
        pad Padding to maintain alignment  This field is reserved for future use
        and must be emitted as zeroes.
        offset  Offset in segment where the code starts
        cbSeg   Count of the number of bytes of code in the segment
*/
void DumpSSTModule(char *data)
{
    struct tagsstModule *s;
    char *p, *st;
    int i;
    s = (struct tagsstModule *) data;
    p = (char *) data;
    p += 8 + (s->cSeg * 12);
    DisplayLengthPrefixedString(p);
    printf(": OvlNr %d idxLib %d Segments %d\n", s->ovlNumber,
           s->iLib, s->cSeg);
    p = (char *) (&data[8]);
    st = (char *) &s->Style;
    for (i = 0; i < s->cSeg; i++) {
        printf("\tSegment %d offset %ld, size %ld Style:'%c%c'\n",
               *(unsigned short *) (&p[0]), *(unsigned long *) (&p[4]), *(unsigned long *) (&p[8]), st[0], st[1]);
        p += 12;
    }
}
/*
(0x0127) sstSrcModule
This table describes the source line number to addressing mapping information for
a module.The table permits the description of a module containing multiple source
files with each source file contributing code to one or more code segments.
The base addresses of the tables described below are all relative to the beginning
of the sstSrcModule table.

Module header
Information for source file 1
Information for segment 1
Information for segment 2
.
Information for source file 2
Information for segment 1
Information for segment 2
.
.
The module header structure describes the source file and code segment organization of the module.
2       2       4*cFile     8*cSeg      2*cSeg
cFile   cSeg    baseSrcFile start/end   seg
cFile   The number of source files contributing code to segments
cSeg    The number of code segments receiving code from this module
baseSrcFile This is an array of base offsets from the beginning of
        the sstSrcModule table
start/end   An array of two 32-bit offsets per segment that receives code from
        this module.  The first offset is the offset within the segment of the
        first byte of code from this module.  The second  offset is the ending
        address of the code from this module.  The order of these pairs corresponds
        to the ordering of the segments in the seg array.  Zeroes in these entries
        means that the information is not known and the file and line tables
        described below need to be examined to determine if an address of interest
        is contained within the code from this module.
seg     An array of segment indices that receive code from this module.  If the
        number of segments is not even, two pad characters are inserted to maintain
        natural alignment.
The file table describes the code segments that receive code from each source file.
2       2   4*cSeg      8*cSeg      2       *
cSeg    pad baseSrcLn   start/end   cbName  Name
cSeg    Number of segments that receive code from this source file.
        If the source file contributes code multiple times to a segment, this is
        reflected in this count.
pad Pad field used to maintain alignment.  This field is reserved for future use
        and must be emitted as zero.
baseSrcLn   An array of offsets for the line/address mapping tables for each of
        the segments that receive code from this source file.
start/end   An array of two 32-bit offsets per segment that receives code from this
        module.  The first offset is the offset within the segment of the first
        byte of code from this module.  The second  offset is the ending address of
        the code from this module.  The order of these pairs corresponds to the
        ordering of the segments in the seg array.  Zeroes in these entries means
        that the information is not known and the file and line tables described
        below need to be examined to determine if an address of interest is
        contained within the code from this module.
cbName  Count of the number of bytes in source file name
Name    Source file name.  This can be a fully or partially qualified path name.

The line number to address mapping information is contained in a table with the
following format.  The preferred ordering for this table is by offset order.
Line number and offsets must be unique.
2   2       4*cPair 2*cPair
Seg cPair   offset  linenumber
Seg     Segment index for this table
cPair   Count of the number of source line pairs to follow
offset  An array of 32-bit offsets for the offset within the code segment of the
        start of the line contained in the parallel array linenumber.
linenumber  This is an array of 16-bit line numbers for the numbers of the lines
        in the source file that cause code to be emitted to the code segment.
        This array is parallel to the offset array.  If cPair is not even, then a
        zero word is emitted to maintain natural alignment in the sstSrcModule table.
*/
SEGMENT_DEBUG_INFO *DumpSSTSrcModule(unsigned char *data, int idx)
{
    int NrOfModules, cSeg, i, cbName, baseSrcLineNumbers, count;
    char *start;
    SEGMENT_DEBUG_INFO *dbgInfoNew;
    printf("sstSrcModule header:\n");
    NrOfModules = *(unsigned short *) data;
    cSeg = *(unsigned short *) data;
    start = data;
    data += 4;
    dbgInfoNew = (SEGMENT_DEBUG_INFO *) malloc(sizeof(SEGMENT_DEBUG_INFO));
    memset(dbgInfoNew, 0, sizeof(SEGMENT_DEBUG_INFO));
    dbgInfoNew->Index = idx;
    printf("Number of src files %d segments %d\n", NrOfModules, cSeg);
    for (i = 0; i < NrOfModules; i++) {
        printf("Base offset %ld\n", *(unsigned long *) (data));
        data += 4;
    }
    for (i = 0; i < cSeg; i++) {
        dbgInfoNew->StartOffset = *(unsigned long *) (&data[0]);
        dbgInfoNew->EndOffset = *(unsigned long *) (&data[4]);
        printf("\tStart %ld End %ld\n",
               *(unsigned long *) (data), *(unsigned long *) (&data[4]));
        data += 8;
    }
    for (i = 0; i < cSeg; i++) {
        printf("\tSegment %d\n", *(unsigned short *) (data));
        data += 2;
    }
    if (cSeg & 1)
        data += 2;
    printf("File table:\n");
    printf("Number of segments %d\n", *(unsigned short *) data);
    data += 4;          /* padding */
    baseSrcLineNumbers = *(unsigned long *) data;
    printf("Base src line numbers %d\n", baseSrcLineNumbers);
    data += 4;
    printf("Code start %ld\t", *(unsigned long *) data);
    data += 4;
    printf("Code end: %ld\n", *(unsigned long *) data);
    data += 4;
    cbName = *data++;
    printf("Module name: ");
    dbgInfoNew->Name = malloc(cbName + 1);
    for (i = 0; i < cbName; i++) {
        dbgInfoNew->Name[i] = *data++;
    }
    dbgInfoNew->Name[i] = 0;
    printf("%s\n", dbgInfoNew->Name);
    data = start + baseSrcLineNumbers;
    printf("Line numbers: seg=%d ", *(unsigned short *) data);
    data += 2;
    count = *(unsigned short *) data;
    dbgInfoNew->LineNrCount = count;
    dbgInfoNew->LinesOffsets = (int *) malloc(count * sizeof(int));
    dbgInfoNew->LinesNrs = (int *) malloc(count * sizeof(int));
    data += 2;
    printf("count =%d\n", count);
    for (i = 0; i < count; i++) {
        dbgInfoNew->LinesOffsets[i] = *(unsigned long *) data;
        data += 4;
    }
    for (i = 0; i < count; i++) {
        dbgInfoNew->LinesNrs[i] = *(unsigned short *) data;
        data += 2;
    }
    for (i = 0; i < count; i++) {
        if (!(i % 5)) {
            if (i)
                printf("\n");
            printf("[%3d] ", i);
        }
        printf("%07d:%05d,", dbgInfoNew->LinesOffsets[i], dbgInfoNew->LinesNrs[i]);
    }
    printf("\n");
    dbgInfoNew->Next = dbgInfoRoot;
    dbgInfoRoot = dbgInfoNew;
    return (dbgInfoNew);
}
/*
Address sort table (addrhash == 12):
The address sort table is a grouping of logical segments (or sections)
where each symbol reference within the segment/section is sorted by its
segment/section relative offset.
The format of the table is as follows.

2	cSeg(n)	Number of logical segments/sections
2	Alignment	Filler to preserve alignment
4n	Segment Table[n]	Each ulong entry is a file offset from the
                        beginning of the offset table to the first offset
                        item for each segment/section.
4n	Offset Counts[n]	Each ulong entry is the count of items in the
                        offset table for each segment.
8m	Offset Table[m]     Each entry is a pair of dwords.  The first dword
                        is the file offset of the referenced symbol from
                        the beginning of the symbols.  The second dword is
                        the segment/section relative offset of the
                        referenced symbol in memory.
n  = the number of segments/sections
m =  the number of symbols (with addresses) = the number of entries in the
     offset table
*/
void DisplayAddressSortTable(char *p)
{
	int *CountTab,*OffTab,i,n,j,realOffset;
	char *start;

	n = *(unsigned short *)p;
	p += 2+2;
	printf("\nAddress hash tables with %d segments\n",n);
	CountTab = (int *)malloc(n * sizeof(int));
	OffTab = (int *)malloc(n * sizeof(int));
	for (i=0; i<n; i++) {
		OffTab[i] = *(unsigned long *)p;
		p+= 4;
	}
	for (i=0; i<n; i++) {
		CountTab[i] = *(unsigned long *)p;
		p+= 4;
	}
	start = p;
	for (i=0; i<n;i++) {
		realOffset = p - start;
		if (realOffset != OffTab[i]) {
			printf("****** realOffset is %d, should be %d!***\n",realOffset,OffTab[i]);
		}
		if (CountTab[i] > 250) {
			printf("**********Count Table[%d] == %d!!!!!****\n",i,CountTab[i]);
//			continue;
		}
		printf("Segment %d: Offset %d Count %d\n",i+1,OffTab[i],CountTab[i]);
		for (j=0; j<CountTab[i];j++) {
			printf("Symbol %3d Offset of debug info %5ld, offset in memory %6ld\n",
				j+1,*(unsigned long *)p,*(unsigned long *)(&p[4]));
			p += 8;
		}
	}
	free(CountTab);
	free(OffTab);
}


struct hashbucket {
    int count;
    int *data;
};
/*
(0x0129) sstGlobalSym
This subsection contains globally compacted symbols.
The format of the table is a header specifying the symbol and address hash functions,
the length of the symbol information, the length of the symbol hash function data,
and the length of address hash function data.  This is followed by the symbol
information followed by the symbol hash tables, followed by the address hash tables.
When the pack utility writes the sstGlobals subsection,  each symbol is zero-padded
such that the following symbol starts on a long boundary and the length field is
adjusted by the pad count.  Note that symbol and/or address hash data can be
discarded and the globally packed symbols linearly searched.
Hash function index 0 means no hash data.  See Section 7.5 for more information
about the hashing functions.
The header has the following format:
2       2           4           4           4
symhash addrhash    cbSymbol    cbSymHash   cbAddrHash
symhash     Index of the symbol hash function
addrhash    Index of the address hash function
cbSymbol    Count of the number of bytes in the symbol table
cbSymHash   Count of the number of bytes in the symbol hash table
cbAddrHash  Count of the number of bytes in the address hashing table
Starting with the NB09 signature files, the sstGlobalSym table can contain S_ALIGN
symbols to maintain 4k alignment of symbols.  Also, starting with NB09 signature
files, the sstGlobal can contain S_PROCREF and S_DATAREF symbols to global
procedures and to global data symbols that would not otherwise have been globally
packed because of symbol type mismatches.  See Section 2.6 for more information
about the S_PROCREF and S_DATAREF symbols.
*/
void DisplaySSTGlobalSymbols(char *data, int idx, int len)
{
    int cbSymbol, n, i, nrOfSymbols;
    char *p, *stop;
    if (idx == 0x129)
        printf("sstGlobalSym");
    else
        printf("sstStaticSym");
    stop = data + len;
    cbSymbol = *(unsigned long *) (&data[4]);
    if (cbSymbol == 0)
        return;
    printf(": symhash %d addrhash %d cbSymbol %d addrHash %ld, cbaddrHash %ld\n",
           *(unsigned short *) (data), *(unsigned short *) (&data[2]),
     cbSymbol, *(unsigned long *) (&data[8]), *(unsigned long *) (&data[12]));
    DumpCodeViewSymbols(data + 16, cbSymbol, NULL, 0);
    printf("Symbol Hash tables\n------ ---- ------\n");
    p = data + 16 + cbSymbol;
    n = *(unsigned short *) (p);
    printf("N= %d. Hash table offsets:\n", n);
	if (n == 0) return;
    p += 4;
    for (i = 0; i < n; i++) {
        if (p > stop) {
            printf("**** synchronisation error ***\n");
            return;
        }
        printf("[%3d] %5ld ", i, *(unsigned long *) p);
        p += 4;
        if ((!((1 + i) % 4)))
            printf("\n");
    }
    printf("\nBucket counts:\n");
    nrOfSymbols = 0;
    for (i = 0; i < n; i++) {
        printf("[%3d] %5ld ", i, *(unsigned long *) p);
        nrOfSymbols += *(unsigned long *) p;
        p += 4;
        if ((!((1 + i) % 4)))
            printf("\n");
    }
    printf("\nChain table with %d symbols\n", nrOfSymbols);
    for (i = 0; i < nrOfSymbols; i++) {
        printf("[%3d] file offset %ld, checksum %ld\n", i,
               *(unsigned long *) p, *(unsigned long *) (p + 4));
        p += 8;
    }
	if (n > 0) DisplayAddressSortTable(p);
}
/*
This subsection contains the globally compacted public symbols from the
sstPublics.  The format of the table is a header specifying the symbol and
address hash functions, the length of the symbol information, the length of the
symbol hash function data, and the length of address hash function data.  This is
followed by the symbol information followed by the symbol hash tables followed by
the address hash tables.  When the pack utility writes the sstGlobals subsection,
each symbol is zero-padded such that the following symbol starts on a long
boundary and the length field of the symbol is adjusted by the pad count.  Note
that symbol and/or address hash data can be discarded and the globally packed
symbols be linearly searched in low-memory situations.  Hash function index 0
means no hash data.  See Section 7.5 for more information about the hashing
functions.
The header has the following format:

2       2           4           4           4

symhash addrhash    cbSymbol    cbSymHash   cbAddrHash

symhash     Index of the symbol hash function
addrhash    Index of the address hash function
cbSymbol    Count of the number of bytes in the symbol table
cbSymHash   Count of the number of bytes in the symbol hash table
cbAddrHash  Count of the number of bytes in the address hashing table

*/
void DisplaySSTGlobalPub(char *data, int idx)
{
    int cbSymbol, n, i, j, nrOfSymbols, *table,cbAddrHash,cbSymHash;
    char *p;

    printf("sstGlobalPub");
    cbSymbol = *(unsigned long *) (&data[4]);
	cbAddrHash = *(unsigned long *)(&data[12]);
	cbSymHash = *(unsigned long *)(&data[8]);
    printf(": symhash function= %d addrhash function= %d\n     cbSymbol %d addrHash %d, cbaddrHash %d\n",
           *(unsigned short *) (data), *(unsigned short *) (&data[2]),
     cbSymbol, cbSymHash, cbAddrHash);
    nrOfSymbols = DumpCodeViewSymbols(data + 16, cbSymbol, NULL, 0);
    printf("Symbol Hash tables\n------ ---- ------\n");
    p = data + 16 + cbSymbol;
    n = *(unsigned short *) (p);
    printf("N= %d. Hash table offsets:\n", n);
    p += 4;
    for (i = 0; i < n; i++) {
        printf("[%3d] %5ld ", i, *(unsigned long *) p);
        p += 4;
        if ((!((1 + i) % 4)))
            printf("\n");
    }
    printf("\nBucket counts:\n");
    nrOfSymbols = 0;
    table = (int *) malloc(n * sizeof(int));
    for (i = 0; i < n; i++) {
        printf("[%3d] %5ld ", i, *(unsigned long *) p);
        nrOfSymbols += *(unsigned long *) p;
        table[i] = *(unsigned long *) p;
        p += 4;
        if ((!((1 + i) % 4)))
            printf("\n");
    }
    printf("\nChain table with %d symbols\n", nrOfSymbols);
    for (i = 0; i < n; i++) {
        for (j = 0; j < table[i]; j++) {
            printf("[%3d %3d] file offset %ld, checksum %ld\n", i, j,
                   *(unsigned long *) p, *(unsigned long *) (p + 4));
            p += 8;
        }
    }
	DisplayAddressSortTable(p);
}
/*
(0x0133) sstFileIndex
This subsection contains an list of all of the sources files that contribute code
to any module (compiland) in the executable.  File names are partially qualified
relative to the compilation directory.
2       2       2 * cMod    2 * cModules    4 * cRef    *
cMod    cRef    ModStart    cRefCnt         NameRef     Names
cMod    Count of the number of modules in the executable
cRef    Count of the total number of file name references
ModStart    Array of indices into the NameOffset table for each module.  Each index
        is the start of the file name references for each module.
cRefCnt Number of file name references per module
NameRef Array of offsets into the Names table.  For each module, the offset to first
        referenced file name is at NameRef[ModStart] and continues for cRefCnt
        entries.
Names   List of zero terminated file names.  Each file name is partially qualified
        relative to the compilation directory.
*/
void DumpFileNames(char *data, int siz)
{
    int cMod, cRef, i, *NameRef, j, z;
    unsigned short *ModStart, *RefCnt;
    printf("File Names\n");
    cMod = *(unsigned short *) data;
    cRef = *(unsigned short *) (&data[2]);
    data += 4;
    printf("Modules %d, Total references %d\n", cMod, cRef);
    ModStart = (short *) malloc(cMod * sizeof(short));
    for (i = 0; i < cMod; i++) {
        ModStart[i] = *(unsigned short *) data;
        data += 2;
    }
    RefCnt = (short *) malloc(sizeof(short) * cMod);
    for (i = 0; i < cMod; i++) {
        RefCnt[i] = *(unsigned short *) data;
        data += 2;
    }
    NameRef = (int *) malloc(cRef * sizeof(int));
    for (i = 0; i < cRef; i++) {
        NameRef[i] = *(unsigned int *) data;
        data += 4;
    }
    z = 0;
    for (i = 0; i < cMod; i++) {
        for (j = 0; j < RefCnt[i]; j++) {
            int idx;
            char *p;
            idx = NameRef[z];
            z++;
            p = &data[idx];
            printf("[%3d] [%3d] ", i, j);
            DisplayLengthPrefixedString(p);
            printf("\n");
        }
    }
}
// Global Segment Info table flags
typedef struct _sgf {
    unsigned short fRead:1;
    unsigned short fWrite:1;
    unsigned short fExecute:1;
    unsigned short f32Bit:1;
    unsigned short res1:4;
    unsigned short fSel:1;
    unsigned short fAbs:1;
    unsigned short res2:2;
    unsigned short fGroup:1;
    unsigned short res3:3;
} SGF;
void DumpSSTSegMap(char *data, int siz)
{
    char *p = data;
    int cseg, csegLog, flags, ovl, group, frame, segName, className, offset, size, i;
    SGF segFlags;
    printf("sstSegMap\n");
    cseg = *(unsigned short *) p;
    p += 2;
    csegLog = *(unsigned short *) p;
    p += 2;
    printf("%d segments, %d logical segments\n", cseg, csegLog);
    for (i = 0; i < cseg; i++) {
        printf("Segment %d\n", i + 1);
        flags = *(unsigned short *) p;
        p += 2;
        ovl = *(unsigned short *) p;
        p += 2;
        group = *(unsigned short *) p;
        p += 2;
        frame = *(unsigned short *) p;
        p += 2;
        printf("Flags %x, ovl %d group %d frame %d ", flags, ovl, group, frame);
        segName = *(unsigned short *) p;
        p += 2;
        className = *(unsigned short *) p;
        p += 2;
        printf("Seg name idx %d Class name idx %d\n", segName, className);
        offset = *(unsigned long *) p;
        p += 4;
        size = *(unsigned long *) p;
        p += 4;
        printf("Offset %x (%dd), Size %d\n", offset, offset, size);
        memcpy(&segFlags, &flags, 2);
#if 0
        if (flags & 1) {
            printf("Readable\n");
        }
        if (flags & 2) {
            printf("Writable\n");
        }
        if (flags & 4) {
            printf("Executable\n");
        }
        if (flags & 8) {
            printf("32 bit\n");
        }
        if (flags & (1 << 8)) {
            printf("Selector\n");
        }
        if (flags & (1 << 9)) {
            printf("Absolute address\n");
        }
        if (flags & (1 << 12)) {
            printf("Group\n");
        }
#else
        if (segFlags.fRead)
            printf("Readable\n");
        if (segFlags.fWrite)
            printf("Writeable\n");
        if (segFlags.fExecute)
            printf("Executable\n");
        if (segFlags.f32Bit)
            printf("32 bit\n");
        if (segFlags.fSel)
            printf("Selector\n");
        if (segFlags.fAbs)
            printf("Absolute address\n");
        if (segFlags.fGroup)
            printf("Group\n");
#endif
    }
}
/*
(0x0128) sstLibraries
There can be at most one sstLibraries Subsection.  The format is an array of
length-prefixed names.  These names define all the library files used during
linking.  The order of this list defines the library index number (see sstModules
subsection).  The first entry should be empty (i.e., zero-length string) since
library indices are 1-based.
*/
void DumpLibraries(char *data, int siz)
{
    int i = 1, j = 1;
    printf("Library Names\n");
    data++;             /* ignore first empty name */
    j = 1;
    while (i < siz) {
        printf("[%3d] '", j);
        DisplayLengthPrefixedString(data);
        i += *data + 1;
        data += *data + 1;
        printf("\n");
        j++;
    }
}
void DumpTypes(char *data, int len)
{
    int n, i;
    long *lp;
    n = *(unsigned long *) (&data[4]);
    printf("Types record with %d types\n", n);
    lp = (unsigned long *) (data + 8);
    for (i = 0; i < n; i++) {
        printf("Type %4d Offset %ld\n", 1 + i, lp[i]);
    }
    data += 8 + n * sizeof(int);
    len -= 8 + n * sizeof(int);
    DumpCodeViewTypes(data, len, 0, NULL);
}
/*
The subsection directory has the format
Directory header
Directory entry 0
Directory entry 1
.
.
.
Directory entry n
The subsection directory is prefixed with a directory header structure indicating size and number of subsection directory entries that follow.
2           2           4       4           4
cbDirHeader cbDirEntry  cDir    lfoNextDir  flags
cbDirHeader Length of the directory header
cbDirEntry  Length of each directory entry
cDir        Number of directory entries
lfoNextDir  Offset from lfaBase of next directory.  This field is currently unused.
            It is intended for use by the incremental linker to point to the next
            directory containing Symbol and Type OMF information from an incremental
            link.
flags       Flags describing directory and subsection tables.  No values have been defined for this field.

The directory header structure is  followed by the directory entries which specify the subsection type, module index if applicable, the subsection offset, and subsection size.
2           2       4   4
subsection  iMod    lfo cb
subsection  Subdirectory index.  See the table below for a listing of the valid
            subsection indices.
iMod        Module index.  This number is 1 based and zero is never a valid index.
            The index 0xffff is reserved for tables that are not associated with a
            specific module.  These tables include sstLibraries, sstGlobalSym,
            sstGlobalPub and sstGlobalTypes.
lfo         Offset from the base address lfaBase
cb          Number of bytes in subsection

There is no requirement for a particular subsection to exist for a particular module
There is a preferred order for subsections within the Symbol and Type OMF portion
and the subsection directory of the file as emitted by the linker (NB05 signature).
The preferred order is:

sstModule1  Module 1
.   .
sstModulen  Module n
sstTypes1   Module 1
sstPublics1 Module 1
sstSymbols1 Module 1
sstSrcModule1   Module 1
.
sstTypesn   Module n
sstPublicsn Module n
sstSymbolsn Module n
sstSrcModulen   Module n
sstLibraries
directory

However, if the tables are not written in this order by the linker, CVPACK will
sort the subsection table into this order and read the subsections in this order by
seeking to the correct location.  The net effect is that packing will be less
efficient but will work.
CVPACK will write the Symbol and Type OMF back to the file in the following order.
CodeView requires that the sstModule entries be first and sequential in the
subsection directory.  For performance reasons, it is recommended that the order of
the subsections in the file match the order of the subsection directory entries.

For signatures prior to NB09, the packed file has the following subsections and
ordering.
NBxx    signature
lfoDir  directory offset
sstModule1  Module 1
.   .
sstModulen  Module n
sstAlignSym1    Module 1
sstSrcModule1   Module 1
.
sstAlignSymn    Module n
sstSrcModulen   Module n
sstGlobalPub    Global Publics
sstGlobalSym    Global Symbols
sstLibraries    Libraries
sstGlobalTypes  Global Types
Directory
NBxx    signature if OMF executable
lfoBase offset of base if OMF executable
For NB09 signatures, the packed file has the following subsections and ordering.
NBxx    signature
lfoDir  directory offset
sstModule1  Module 1
.   .
sstModulen  Module n
sstAlignSym1    Module 1
sstSrcModule1   Module 1
.
sstAlignSymn    Module n
sstSrcModulen   Module n
sstGlobalPub    Global Publics
sstGlobalSym    Global Symbols
sstLibraries    Libraries
sstGlobalTypes  Global Types
sstStaticSym    Static Symbols
sstFileIndex    File Index
sstOffsetMap16 sstOffsetMat32
Directory
NBxx    signature
lfoBase offset

*/
void DumpSSEntry(char *data, char *pBase)
{
    struct subsection *s;
    static SEGMENT_DEBUG_INFO *sdi;
    s = (struct subsection *) data;
    if (s->subsection ||s->iMod ||s->ifo ||s->cb)
    printf("Subsection %x Module %d offset %ld with %ld bytes\n",
           s->subsection, s->iMod, s->ifo, s->cb);
    switch (s->subsection) {
    case sstModule:
        DumpSSTModule(pBase + s->ifo);
        sdi = NULL;
        break;
    case 0x127:
        sdi = DumpSSTSrcModule(pBase + s->ifo, s->iMod);
        break;
    case 0x125:
        DumpCodeViewSymbols(pBase + s->ifo, s->cb, sdi, 1);
        break;
    case 0x128:
        DumpLibraries(pBase + s->ifo, s->cb);
        break;
    case 0x12b:
        DumpTypes(pBase + s->ifo, s->cb);
        break;
    case 0x12d:
        DumpSSTSegMap(pBase + s->ifo, s->cb);
        break;
    case 0x133:
        DumpFileNames(pBase + s->ifo, s->cb);
        break;
    case 0x129:
    case 0x134:
        DisplaySSTGlobalSymbols(pBase + s->ifo, s->subsection, s->cb);
        break;
    case 0x12a:
        DisplaySSTGlobalPub(pBase + s->ifo, s->subsection);
        break;
    }
}
/*
The subsection directory is prefixed with a directory header structure indicating
size and number of subsection directory entries that follow.
2           2           4       4           4
cbDirHeader cbDirEntry  cDir    lfoNextDir  flags
cbDirHeader Length of the directory header
cbDirEntry  Length of each directory entry
cDir    Number of directory entries
lfoNextDir  Offset from lfaBase of next directory.  This field is currently unused.   It is intended for use by the incremental linker to point to the next directory containing Symbol and Type OMF information from an incremental link.
flags   Flags describing directory and subsection tables.  No values have been defined for this field.
*/
void DumpSSDir(char *p, char *pBase)
{
    unsigned short cbDirHeader, cbDirEntry;
    unsigned long cDir, IfoNextDir, flags;
    unsigned int i;
    cbDirHeader = *(unsigned short *) (p);
    cbDirEntry = *(unsigned short *) (&p[2]);
    cDir = *(unsigned long *) (&p[4]);
    IfoNextDir = *(unsigned long *) (&p[8]);
    flags = *(unsigned long *) (&p[12]);
    printf("Header size %d, %ld entries of %d bytes. Next %ld flags %ld\n",
           cbDirHeader, cDir, cbDirEntry, IfoNextDir, flags);
    p += cbDirHeader;
    for (i = 0; i < cDir; i++) {
        DumpSSEntry(p, pBase);
        p += cbDirEntry;
    }
}
/*
The debug information format encompasses a block of data which goes into the .exe
file at a location dependent upon the executable file format.  The version of the
debug information is specified by a signature that is contained within the debug
information.  The signature has the format  NBxx, where xx is the version number
and have the following meanings.
NB00    Not supported
NB01    Not supported
NB02    Linked by a Microsoft LINK, version 5.10, or equivalent OEM linker
NB03    Not supported
NB04    Not supported
NB05    The signature emitted by LINK 5.20 and later linkers for a file before it
        has been packed.
NB06    Not supported.
NB07    This signature is used for Quick C for Windows 1.0 only.
NB08    The signature CodeView 4.00 through CodeView 4.05 uses for a file after it has been packed.  CodeView 4.00 - 4.05 will not process a file that does not have this signature.
NB09    The signature for CodeView 4.10 for a file after it has been packed.
        CodeView 4.10 will not process a file that does not have this signature.
The method for finding the debug information depends upon the executable format.
PE Format:
For PE format executables, the base address lfaBase is found by examining the
executable header.  Note: temporarily, MS code uses the same method to find the
debug information as is used for OMF format executables.
executable header   Contains pointer to debug information
executable code + ...
NBxx    Signature at lfaBase
lfoDirectory    Offset of directory from base address (lfoDir)
Subsection tables   sstModule, sstType, sstLibraries, ...
.
.
.
Subsection Directory    At file offset lfaBase + lfoDir
other information
All other file offsets in the Symbol and Type OMF are relative to lfaBase.  At the
base address the signature is repeated, followed by the long displacement to the
subsection directory (lfoDir).  All subsections start on a long word boundary and
are designed to maintain natural alignment internally in each subsection and within
the subsection directory.
*/
void DumpExeCodeViewInfo(DWORD base, PIMAGE_DEBUG_DIRECTORY debugDir)
{
    char *p, signature[6], *pBase;
    int i;
    DWORD ssDir;
    pBase = MakePtr(PCHAR, base, debugDir->PointerToRawData);
    for (i = 0; i < 4; i++) {
        signature[i] = pBase[i];
    }
    signature[i] = 0;
    ssDir = *(unsigned long *) (&pBase[4]);
    printf("Signature: %s, IfoDir Offset: %ld\n", signature, ssDir);
    p = MakePtr(PCHAR, base, debugDir->PointerToRawData + ssDir);
    DumpSSDir(p, pBase);
}
/*
 Dump the debug directory in a PE file.

6.1.1. Debug Directory (Image Only)
--------------------------
 Image files contain an optional 'debug directory' indicating what form of
 debug information is present and where it is. This directory consists of
 an array of 'debug directory entries' whose location and size are
 indicated in the image optional header. The debug directory may be in a
 discardable .debug section (if one exists) or it may be included in any
 other section in the image file, or not in a section at all. Each debug
 directory entry identifies the location and size of a block of debug
 information. The RVA specified may be 0 if the debug information is not
 covered by a section header (i.e., it resides in the image file and is not
 mapped into the run-time address space). If it is mapped, the RVA is its
 address. Here is the format of a debug directory entry:
 Offset  Size  Field              Description
 0       4     Characteristics	  A reserved field intended to be used for
                                  flags, set to zero for now.
 4       4     Time/Date Stamp    Time and date the debug data was created.
 8       2     Major Version      Major version number of the debug data format.
 10      2     Minor Version      Minor version number of the debug data format.
 12      4	   Debug Type         Format of debugging information: this
                                  field enables support of multiple
                                  debuggers.
 16      4     Size of Data       Size of the debug data (not including the
                                  debug directory itself).
 20      4     RVA of Raw Data    Address of the debug data when loaded,
                                  relative to the image base.
 24      4     Pointer to Raw Data File pointer to the debug data.

The following values are defined  for the Debug Type field of the debug
directory:
Constant                Value   Description
IMAGE_DEBUG_TYPE_UNKOWN 0       Unknown value, ignored by all tools.
IMAGE_DEBUG_TYPE_COFF   1       COFF debug information (line numbers,
                                symbol table, and string table). This type
                                of debug information is also pointed to by
                                fields in the file headers.
IMAGE_DEBUG_TYPE_CODEVIEW 2     CodeView debug information.
IMAGE_DEBUG_TYPE_FPO      3     Frame Pointer Omission (FPO) information.
                                This information tells the debugger how to
                                interpret non-standard stack frames, which
                                use the EBP register for a purpose other
                                than as a frame pointer. If Debug Type is
                                set to IMAGE_DEBUG_TYPE_FPO, the debug raw
                                data is an array in which each member
                                describes the stack frame of a function.
                                Not every function in the image file need
                                have FPO information defined for it, even
                                though debug type is FPO. Those functions
                                that do not have FPO information are
                                assumed to have normal stack frames. The
                                format for FPO information is defined as
                                follows:

#define FRAME_FPO   0
#define FRAME_TRAP  1
#define FRAME_TSS   2

typedef struct _FPO_DATA {
DWORD       ulOffStart; // offset 1st byte of function code
DWORD       cbProcSize;            // # bytes in function
DWORD       cdwLocals;             // # bytes in locals/4
WORD        cdwParams;             // # bytes in params/4
WORD        cbProlog : 8;          // # bytes in prolog
WORD        cbRegs   : 3;          // # regs saved
WORD        fHasSEH  : 1;          // TRUE if SEH in func
WORD        fUseBP   : 1;          // TRUE if EBP has been allocated
WORD        reserved : 1;          // reserved for future use
WORD        cbFrame  : 2;          // frame type
 } FPO_DATA;
 */
PIMAGE_DEBUG_DIRECTORY debugDir;
void DumpDebugDirectory(DWORD base, PIMAGE_NT_HEADERS pNTHeader)
{
    PIMAGE_SECTION_HEADER header;
    unsigned cDebugFormats, i;
    DWORD offsetInto_rdata;
    DWORD va_debug_dir;
    PSTR szDebugFormat;
    // This line was so long that we had to break it up
    va_debug_dir = pNTHeader->OptionalHeader.
        DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].
        VirtualAddress;
    if (va_debug_dir == 0)
        return;
    // If we found a .debug section, and the debug directory is at the
    // beginning of this section, it looks like a Borland file
    header = GetSectionHeader(".debug", pNTHeader);
    if (header && (header->VirtualAddress == va_debug_dir)) {
        debugDir = (PIMAGE_DEBUG_DIRECTORY) (header->PointerToRawData + base);
        cDebugFormats = pNTHeader->OptionalHeader.
            DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size;
    }
    else {              // Look for microsoft debug directory in the .rdata section
        header = GetSectionHeader(".rdata", pNTHeader);
        if (!header)
            return;
        // See if there's even any debug directories to speak of...
        cDebugFormats = pNTHeader->OptionalHeader.
            DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size
            / sizeof(IMAGE_DEBUG_DIRECTORY);
        if (cDebugFormats == 0)
            return;
        offsetInto_rdata = va_debug_dir - header->VirtualAddress;
        debugDir = MakePtr(PIMAGE_DEBUG_DIRECTORY, base,
                           header->PointerToRawData + offsetInto_rdata);
    }
    printf(
           "Debug Formats in File\n"
    "  Type            Size     Address  FilePtr  Charactr TimeData Version\n"
    "  --------------- -------- -------- -------- -------- -------- --------\n"
        );
    for (i = 0; i < cDebugFormats; i++) {
        szDebugFormat = (debugDir->Type <= 6)
            ? SzDebugFormats[debugDir->Type] : "???";
        printf("  %-15s %08ld %08ld %08ld %08ld %08ld %u.%02u\n",
               szDebugFormat, debugDir->SizeOfData, debugDir->AddressOfRawData,
               debugDir->PointerToRawData, debugDir->Characteristics,
               debugDir->TimeDateStamp, debugDir->MajorVersion,
               debugDir->MinorVersion);
        // If COFF debug info, save its address away for later.  We
        // do the check for "PointerToSymbolTable" because some files
        // have bogus values for the COFF header offset.
        if ((debugDir->Type == IMAGE_DEBUG_TYPE_COFF) &&
                pNTHeader->FileHeader.PointerToSymbolTable) {
            PCOFFDebugInfo =
                (PIMAGE_DEBUG_INFO) (base + debugDir->PointerToRawData);
        }
        if (fDebug && debugDir->Type == 2 && debugDir->SizeOfData)  /* Code
                         view debug info */
            DumpExeCodeViewInfo(base, debugDir);
        debugDir++;
    }
}
// Function prototype (necessary because two functions recurse)
void DumpResourceDirectory
(
  PIMAGE_RESOURCE_DIRECTORY resDir, DWORD resourceBase,
  DWORD level, DWORD resourceType
);
// The predefined resource types
char *SzResourceTypes[] = {
    "???_0", "CURSOR", "BITMAP", "ICON", "MENU", "DIALOG", "STRING", "FONTDIR",
    "FONT", "ACCELERATORS", "RCDATA", "MESSAGETABLE", "GROUP_CURSOR",
    "???_13", "GROUP_ICON", "???_15", "VERSION"
};
// Get an ASCII string representing a resource type
void GetResourceTypeName(DWORD type, PSTR buffer, UINT cBytes)
{
    if (type <= 16)
        strncpy(buffer, SzResourceTypes[type], cBytes);
    else
        sprintf(buffer, "%ld", type);
}
//
                        // If a resource entry has a string name (rather than an ID), go find
                        // the string and convert it from unicode to ascii.
                        //
void GetResourceNameFromId(DWORD id, DWORD resourceBase, PSTR buffer, UINT cBytes)
{
    PIMAGE_RESOURCE_DIR_STRING_U prdsu;
    // If it's a regular ID, just format it.
    if (!(id & IMAGE_RESOURCE_NAME_IS_STRING)) {
        sprintf(buffer, "%ld", id);
        return;
    }
    id &= 0x7FFFFFFF;
    prdsu = (PIMAGE_RESOURCE_DIR_STRING_U) (resourceBase + id);
    // prdsu->Length is the number of unicode characters
    WideCharToMultiByte(CP_ACP, 0, prdsu->NameString, prdsu->Length,
                        buffer, 50, 0, 0);
    buffer[min(cBytes - 1, prdsu->Length)] = 0; // Null terminate it!!!
    if (strlen(buffer) < (cBytes - 6)) {
        sprintf(buffer + strlen(buffer), ":%ld", id);
    }
}
typedef struct {
    DWORD DataSize;     // Size of data.
    DWORD HeaderSize;   // Size of the resource header.
} RES, *PRES;
void DumpOneResource(int level, PIMAGE_RESOURCE_DIRECTORY_ENTRY resDirEntry, DWORD resourceBase)
{
    int i, VirtAddress;
    char *data;
    WORD *pword;
    for (i = 0; i < level; i++)
        printf("    ");
#ifdef __LCC__
    data = (char *) (resDirEntry->u2.OffsetToData & 0x7FFFFFFF);
#else
    data = (char *) (resDirEntry->OffsetToData & 0x7FFFFFFF);
#endif
    data += resourceBase;
    pword = (WORD *) data;
    if (resSection)
        VirtAddress = resSection->VirtualAddress;
    else
        VirtAddress = 0;
    printf("Word 1: %d Word 2: %d long 1:%ld long 0:%ld\n",
           pword[0], pword[1], *(unsigned long *) data - VirtAddress,
           *(unsigned long *) pword);
}
//
// Dump the information about one resource directory entry.  If the
// entry is for a subdirectory, call the directory dumping routine
// instead of printing information in this routine.
//
void DumpResourceEntry(PIMAGE_RESOURCE_DIRECTORY_ENTRY resDirEntry,
                       DWORD resourceBase, DWORD level)
{
    UINT i, fileOffset,off,name;
    char nameBuffer[128];

#ifdef __LCC__
	off = resDirEntry->u2.OffsetToData;
	name = resDirEntry->u1.Name;
#else
	name = resDirEntry->Name;
	off = resDirEntry->OffsetToData;
#endif
	i = off & IMAGE_RESOURCE_DATA_IS_DIRECTORY;

    if (i) {
        DumpResourceDirectory((PIMAGE_RESOURCE_DIRECTORY)
                    ((off & 0x7FFFFFFF) + resourceBase),
                              resourceBase, level, name);
        return;
    }
    // Spit out the spacing for the level indentation
    for (i = 0; i < level; i++)
        printf("    ");
    fileOffset = (DWORD) resDirEntry - resourceBase;
    if (name & IMAGE_RESOURCE_NAME_IS_STRING) {
        GetResourceNameFromId(name, resourceBase, nameBuffer,
                              sizeof(nameBuffer));
        printf("Name at %d:%d %s  Offset: %08d\n", fileOffset, name & 0x7FFFFFFF,
               nameBuffer, off);
    }
    else {
        printf("ID: at %d: %08d  Offset: %08d\n", fileOffset,
               name, off);
        DumpOneResource(level, resDirEntry, resourceBase);
    }
}
//
// Dump the information about one resource directory.
//
void DumpResourceDirectory(
                           PIMAGE_RESOURCE_DIRECTORY resDir,
                           DWORD resourceBase,
                           DWORD level,
                           DWORD resourceType
)
{
    PIMAGE_RESOURCE_DIRECTORY_ENTRY resDirEntry;
    char szType[64];
    UINT i, fileOffset;
    // Spit out the spacing for the level indentation
    for (i = 0; i < level; i++)
        printf("    ");
    // Level 1 resources are the resource types
    if (level == 1 && !(resourceType & IMAGE_RESOURCE_NAME_IS_STRING)) {
        GetResourceTypeName(resourceType, szType, sizeof(szType));
    }
    else {              // Just print out the regular id or name
        GetResourceNameFromId(resourceType, resourceBase, szType,
                              sizeof(szType));
    }
    fileOffset = (DWORD) resDir - resourceBase;
    printf(
        "ResDir (%s) at %d. Named:%2d ID:%2d Time:%08lX Vers:%u.%2u Char:%lX\n",
           szType, fileOffset, resDir->NumberOfNamedEntries, resDir->NumberOfIdEntries,
           resDir->TimeDateStamp, resDir->MajorVersion,
           resDir->MinorVersion, resDir->Characteristics);
    resDirEntry = (PIMAGE_RESOURCE_DIRECTORY_ENTRY) (resDir + 1);
    for (i = 0; i < resDir->NumberOfNamedEntries; i++, resDirEntry++)
        DumpResourceEntry(resDirEntry, resourceBase, level + 1);
    for (i = 0; i < resDir->NumberOfIdEntries; i++, resDirEntry++)
        DumpResourceEntry(resDirEntry, resourceBase, level + 1);
}
//
                        // Top level routine called to dump out the entire resource hierarchy
                        //
void DumpResourceSection(DWORD base, PIMAGE_NT_HEADERS pNTHeader)
{
    PIMAGE_RESOURCE_DIRECTORY resDir = GetSectionPtr(".rsrc", pNTHeader, (DWORD) base);
    if (!resDir)
        return;
    resSection = GetSectionHeader(".rsrc", pNTHeader);
    printf("Resources\n");
    DumpResourceDirectory(resDir, (DWORD) resDir, 0, 0);
}
//
                        // Dump the imports table (the .idata section) of a PE file
                        //
void DumpImportsSection(DWORD base, PIMAGE_NT_HEADERS pNTHeader)
{
    PIMAGE_IMPORT_DESCRIPTOR importDesc;
    PIMAGE_SECTION_HEADER header;
    PIMAGE_THUNK_DATA thunk;
    PIMAGE_IMPORT_BY_NAME pOrdinalName;
    DWORD exportsStartRVA, exportsEndRVA;
    int offset;
    INT delta;
	unsigned long Characteristics;

    header = GetSectionHeader(".idata", pNTHeader);
    if (!header)
        return;
    importDesc = MakePtr(PIMAGE_IMPORT_DESCRIPTOR, base,
                         header->PointerToRawData);
    delta = (INT) (header->VirtualAddress - header->PointerToRawData);
    offset = (unsigned long) importDesc - (unsigned long) base;
    printf("Imports Table size:%ld\n", header->SizeOfRawData);
    while (1) {
        int diff;
#ifdef __LCC__
		Characteristics = (unsigned long)importDesc->u1.Characteristics;
#else
		Characteristics = (unsigned long)importDesc->Characteristics;
#endif
        // See if we've reached an empty IMAGE_IMPORT_DESCRIPTOR
        if ((importDesc->TimeDateStamp == 0) && (importDesc->Name == 0))
            break;
        offset = (unsigned long) importDesc - (unsigned long) base;
        if (IsBadReadPtr(((PBYTE)importDesc->Name)+base-delta, 2)) {
		printf("*** bad read pointer %lx %ld at offset %d\n",
			importDesc->Name,importDesc->Name,offset);
	}
        printf("  offset %d %s\n", offset, (PBYTE) (importDesc->Name) - delta + base);
        printf("  Hint/Name Table: %08lX\n", Characteristics);
        printf("  TimeDateStamp:   %08lX\n", importDesc->TimeDateStamp);
        printf("  ForwarderChain:  %08lX\n", importDesc->ForwarderChain);
        printf("  First thunk RVA: %08lX", (unsigned long)importDesc->FirstThunk);
        if ((diff = (unsigned long) importDesc->FirstThunk - (unsigned long) Characteristics)) {
            printf(" (delta: %d 0x%x) ", diff, diff);
        }
        printf("\n");
        thunk = (PIMAGE_THUNK_DATA) importDesc->FirstThunk;
        thunk = (PIMAGE_THUNK_DATA) ((PBYTE) thunk - delta + base);
        // If the pointer that thunk points to is outside of the .idata
        // section, it looks like this file is "pre-fixed up" with regards
        // to the thunk table.  In this situation, we'll need to fall back
        // to the hint-name (aka, the "Characteristics") table.
        exportsStartRVA = header->VirtualAddress;
        exportsEndRVA = exportsStartRVA + header->SizeOfRawData;
		if (IsBadReadPtr(thunk,4)) {
			printf("*** bad read pointer in exports table!\n");
			break;
		}
        if ((*(PDWORD) thunk <= exportsStartRVA) ||
                (*(PDWORD) thunk >= exportsEndRVA)) {
            if (Characteristics == 0)   // Borland doesn't have
                return; // this table!!!
            thunk = (PIMAGE_THUNK_DATA) Characteristics;
            if (((DWORD) thunk <= exportsStartRVA) ||
                    ((DWORD) thunk >= exportsEndRVA))
                return;
            thunk = (PIMAGE_THUNK_DATA) ((PBYTE) thunk - delta + base);
        }
        printf("  Ordn  Name\n");
        while (1) {     // Loop forever (or until we break out)
            if (!IsGoodPointer(thunk))
                break;
            if (thunk->u1.AddressOfData == 0)
                break;
            if (thunk->u1.Ordinal & IMAGE_ORDINAL_FLAG) {
                printf("  %4lu  <imported by ordinal>\n", thunk->u1.Ordinal & 0xFFFF);
                /*break; Bonnes 971219 Show those undocumented functions too*/
            }
            else {
                pOrdinalName = thunk->u1.AddressOfData;
                pOrdinalName = (PIMAGE_IMPORT_BY_NAME)
                    ((PBYTE) pOrdinalName - delta + base);
                if (!IsGoodPointer(pOrdinalName)) {
                    break;
                }
                printf("  %4u  %s %lx\n", pOrdinalName->Hint, pOrdinalName->Name,
					(DWORD)thunk->u1.AddressOfData);
            }
            thunk++;    // Advance to next thunk
        }
        importDesc++;   // advance to next IMAGE_IMPORT_DESCRIPTOR
        printf("\n");
    }
}
//
                        // Dump the exports table (the .edata section) of a PE file
                        //
extern PIMAGE_SECTION_HEADER pSections;
void DumpExportsSection(DWORD base, PIMAGE_NT_HEADERS pNTHeader)
{
    PIMAGE_EXPORT_DIRECTORY exportDir;
    PIMAGE_SECTION_HEADER header;
    INT delta;
    PSTR filename;
    DWORD i,j;
    PDWORD functions;
    PWORD ordinals;
    PSTR *name;
    char *ptime;
    int NumberOfSections;
	char ptmpbuf[10],**pImportNames;;

    header = GetSectionHeader(".edata", pNTHeader);
    if (!header) {
	DWORD prw = (DWORD)pNTHeader->OptionalHeader.DataDirectory[0].VirtualAddress;
	if (prw) {
		unsigned int i;
		header = pSections;
		NumberOfSections = pNTHeader->FileHeader.NumberOfSections;
		for (i=0; i<NumberOfSections;i++) {
			if ((DWORD)header->VirtualAddress <= prw &&
				((DWORD)header->SizeOfRawData+(DWORD)header->VirtualAddress) > prw)
				break;
			header++;
		}
		if (header == NULL) {
			return;
		}
		strncpy(ptmpbuf,header->Name,8);
		ptmpbuf[8] = 0;
		printf("Export table in %s section offset 0x%lx<=0x%lx<=0x%lx\n",
                        ptmpbuf,header->VirtualAddress,
                        prw,header->VirtualAddress+header->SizeOfRawData);
                prw = (DWORD)header->PointerToRawData +
                        (DWORD)pNTHeader->OptionalHeader.DataDirectory[0].VirtualAddress;
                prw -= header->VirtualAddress;
                exportDir = MakePtr(PIMAGE_EXPORT_DIRECTORY, base,prw);
                delta = (INT)(header->VirtualAddress - header->PointerToRawData);
                pImportNames = (char **)((DWORD)exportDir->AddressOfNames - delta + base);
		goto suite;
	}
	else return;
    }
    exportDir = MakePtr(PIMAGE_EXPORT_DIRECTORY, base,
                        header->PointerToRawData);
    delta = (INT) (header->VirtualAddress - header->PointerToRawData);
suite:
    filename = (PSTR) (exportDir->Name - delta + base);
    printf("exports table:\n\n");
	if (!IsGoodPointer(filename)) {
		return;
	}
    printf("  Name:            %s\n", filename);
    printf("  Characteristics: %08lX\n", exportDir->Characteristics);
    printf("  TimeDateStamp:   %08lX ", (unsigned long)exportDir->TimeDateStamp);
    ptime = ctime(&exportDir->TimeDateStamp);
    printf("%s",ptime);
    printf("  Version:         %u.%02u\n", exportDir->MajorVersion,
           exportDir->MinorVersion);
    printf("  Ordinal base:    %08lX\n", exportDir->Base);
    printf("  # of functions:  %08ld\n", exportDir->NumberOfFunctions);
    printf("  # of Names:      %08ld\n", exportDir->NumberOfNames);
    functions = (PDWORD) ((DWORD) exportDir->AddressOfFunctions - delta + base);
    for (i = exportDir->Base,j=0; i < exportDir->NumberOfFunctions; i++)
       if (*(functions+i)) j++;
    printf("  # of entries:    %08ld\n", j);
    ordinals = (PWORD) ((DWORD) exportDir->AddressOfNameOrdinals - delta + base);
    name = (PSTR *) ((DWORD) exportDir->AddressOfNames - delta + base);
    printf("Adresses of functions: %08lX\nAdresses of name Ordinals %08lX\n"
           "Adresses of names: %08lx\n",
	(unsigned long)exportDir->AddressOfFunctions,(unsigned long)exportDir->AddressOfNameOrdinals,
	(unsigned long)exportDir->AddressOfNames);
    printf("\n  Entry Pt  Ordn  Name\n");
    for (i = exportDir->Base; i < exportDir->NumberOfFunctions; i++) {
      if (*(functions+i)) {
         for (j=0; j < exportDir->NumberOfNames; j++) {
            if ( i == (*(ordinals + j) + exportDir->Base))  /* Entry exported by name*/
            break;
         }
         if (j < exportDir->NumberOfNames ) 
            printf("  %08lX  %4lu  %s\n", *(functions+i),
                 *(ordinals +j) + exportDir->Base,
                 (*(name +j)- delta + base));
         else 
            printf("  %08lX  %4lu <exported by ordinal>\n",*(functions+i),i);
       }
    }
}
// The names of the available base relocations
char *SzRelocTypes[] = {
    "ABSOLUTE", "HIGH", "LOW", "HIGHLOW", "HIGHADJ", "MIPS_JMPADDR",
"I860_BRADDR", "I860_SPLIT"};
//
                        // Dump the base relocation table of a PE file
                        //
void DumpBaseRelocationsSection(DWORD base, PIMAGE_NT_HEADERS pNTHeader)
{
    PIMAGE_BASE_RELOCATION baseReloc;
    int filePosition,size,totalEntries;
    extern void *lpFileBase;
    baseReloc = GetSectionPtr(".reloc", pNTHeader, base);
    if (!baseReloc)
        return;
    printf("base relocations:\n");
	size = 0;
	totalEntries = 0;
    while (baseReloc->SizeOfBlock != 0) {
        unsigned i, cEntries;
        PWORD pEntry;
        char *szRelocType;
        WORD relocType;
        cEntries = (baseReloc->SizeOfBlock - sizeof(*baseReloc)) / sizeof(WORD);
        totalEntries += cEntries;
        pEntry = MakePtr(PWORD, baseReloc, sizeof(*baseReloc));
	filePosition = (DWORD)pEntry - (DWORD)lpFileBase;
        printf("\nVirtual Address: %08lx File Position %d size: %08ld %d entries\n\n",
               baseReloc->VirtualAddress, 
		filePosition,baseReloc->SizeOfBlock,cEntries);
	size += baseReloc->SizeOfBlock;
        for (i = 0; i < cEntries; i++) {
            // Extract the top 4 bits of the relocation entry.  Turn those 4
            // bits into an appropriate descriptive string (szRelocType)
            relocType = (*pEntry & 0xF000) >> 12;
            szRelocType = relocType < 8 ? SzRelocTypes[relocType] : "unknown";
            printf("[%3d] %08ld 0x%08lx %10s \n",i+1,
                   (*pEntry & 0x0FFF) + baseReloc->VirtualAddress+0x400000,
                   (*pEntry & 0x0FFF) + baseReloc->VirtualAddress+0x400000,
                   szRelocType);
            pEntry++;   // Advance to next relocation entry
        }
        baseReloc = MakePtr(PIMAGE_BASE_RELOCATION, baseReloc,
                            baseReloc->SizeOfBlock);
    }
    printf("\nTotal size %d, total number of entries: %d\n",size+8,totalEntries);
}
//
                        // Dump the COFF debug information header
                        //
void DumpCOFFHeader(PIMAGE_DEBUG_INFO pDbgInfo)
{
    printf("COFF Debug Info Header\n");
    printf("  NumberOfSymbols:      %08ld\n", pDbgInfo->NumberOfSymbols);
    printf("  LvaToFirstSymbol:     %08ld\n", pDbgInfo->LvaToFirstSymbol);
    printf("  NumberOfLinenumbers:  %08ld\n", pDbgInfo->NumberOfLinenumbers);
    printf("  LvaToFirstLinenumber: %08ld\n", pDbgInfo->LvaToFirstLinenumber);
    printf("  RvaToFirstByteOfCode: %08ld\n", pDbgInfo->RvaToFirstByteOfCode);
    printf("  RvaToLastByteOfCode:  %08ld\n", pDbgInfo->RvaToLastByteOfCode);
    printf("  RvaToFirstByteOfData: %08ld\n", pDbgInfo->RvaToFirstByteOfData);
    printf("  RvaToLastByteOfData:  %08ld\n", pDbgInfo->RvaToLastByteOfData);
}
/*
The MS-DOS Stub is a valid application that runs under MS-DOS and is placed at the
front of the .EXE image. The linker places a default stub here, which prints out
the message "This program cannot be run in DOS mode" when the image is run in MS-DOS
The user can specify another stub by using the /STUB linker option.
At location 0x3c, the stub has the file offset to the Portable Executable (PE)
signature. This information enables Windows NT to properly execute the image file,
even though it has a DOS Stub. This file offset is placed at location 0x3c during
linking.
*/
void DumpExeFile(PIMAGE_DOS_HEADER dosHeader)
{
    PIMAGE_NT_HEADERS pNTHeader;
    DWORD base = (DWORD) dosHeader;
    extern PIMAGE_SECTION_HEADER pSections;
    pNTHeader = MakePtr(PIMAGE_NT_HEADERS, dosHeader,
                        dosHeader->e_lfanew);
    // First, verify that the e_lfanew field gave us a reasonable
    // pointer, then verify the PE signature.
    if (IsBadReadPtr(pNTHeader, sizeof(IMAGE_NT_HEADERS)) ||
            pNTHeader->Signature != IMAGE_NT_SIGNATURE) {
        printf("Unhandled EXE type, or invalid .EXE\n");
        return;
    }
    DumpHeader((PIMAGE_FILE_HEADER) & pNTHeader->FileHeader);
    printf("\n");
    if (!fExtract) {
        DumpOptionalHeader((PIMAGE_OPTIONAL_HEADER) & pNTHeader->OptionalHeader);
        printf("\n");
    }
    DumpSectionTable((PIMAGE_SECTION_HEADER) (pNTHeader + 1),
                     pNTHeader->FileHeader.NumberOfSections, TRUE);
    printf("\n");
    pSections = (PIMAGE_SECTION_HEADER) (pNTHeader + 1);
    if (!fExtract) {
        DumpDebugDirectory(base, pNTHeader);
        printf("\n");
    }
    else
        return;
    if (fResources) {
        DumpResourceSection(base, pNTHeader);
        printf("\n");
    }
    DumpImportsSection(base, pNTHeader);
    printf("\n");
    DumpExportsSection(base, pNTHeader);
    printf("\n");
    if (fShowRelocations) {
        DumpBaseRelocationsSection(base, pNTHeader);
        printf("\n");
    }
    //
    // Initialize these vars here since we'll need them in DumpLineNumbers
    //
    PCOFFSymbolTable = MakePtr(PIMAGE_SYMBOL, base,
                               pNTHeader->FileHeader.PointerToSymbolTable);
    COFFSymbolCount = pNTHeader->FileHeader.NumberOfSymbols;
    if (PCOFFDebugInfo) {
        DumpCOFFHeader(PCOFFDebugInfo);
        printf("\n");
    }
    if (fShowLineNumbers && PCOFFDebugInfo) {
        DumpLineNumbers(MakePtr(PIMAGE_LINENUMBER, PCOFFDebugInfo,
                                PCOFFDebugInfo->LvaToFirstLinenumber),
                        PCOFFDebugInfo->NumberOfLinenumbers, (DWORD) PCOFFDebugInfo - (DWORD) dosHeader + PCOFFDebugInfo->LvaToFirstLinenumber);
        printf("\n");
    }
    if (fShowSymbolTable) {
        if (pNTHeader->FileHeader.NumberOfSymbols
                && pNTHeader->FileHeader.PointerToSymbolTable) {
            DumpSymbolTable(PCOFFSymbolTable, COFFSymbolCount);
            printf("\n");
        }
    }
    if (fShowRawSectionData) {
        DumpRawSectionData((PIMAGE_SECTION_HEADER) (pNTHeader + 1),
                           dosHeader,
                           pNTHeader->FileHeader.NumberOfSections);
    }
}
extern char SourceFileName[];
int FindExeLineNrs(unsigned int idx)
{
    SEGMENT_DEBUG_INFO *rvp;
    int i, count;
    rvp = dbgInfoRoot;
    while (rvp) {
        if (idx <= rvp->EndOffset && idx >= rvp->StartOffset) {
            count = rvp->LineNrCount;
            if (idx == rvp->StartOffset) {
                strcpy(SourceFileName, rvp->Name);
            }
            for (i = 0; i < count; i++) {
                if ((unsigned) (rvp->LinesOffsets[i]) == idx) {
                    if (i == (count - 1)) {
                        *SourceFileName = 0;
                    }
                    printf("Line %d %s\n", rvp->LinesNrs[i], rvp->Name);
                }
            }
        }
        rvp = rvp->Next;
    }
    return (0);
}
